#pragma once
//==========================================================================
// MathUtil.h
//   w֌W̏֐NX
//==========================================================================

#include "Util.h"

//------------------------------------------------------------
// dcomp
//   fl^
// polygon
//   v/v,dȂ̓_őp`\
//------------------------------------------------------------

typedef complex<double> dcomp;
typedef vector<dcomp>   polygon;

//----------------------------------------------------------------
// 덷lr
//----------------------------------------------------------------

static const double EPS = numeric_limits<double>::epsilon();

inline bool is_zero( double a ) { return 0-EPS<=a && a<=0+EPS; }
inline bool is_one ( double a ) { return 1-EPS<=a && a<=1+EPS; }
inline bool is_ordered( double a, double b, double c ) { return a-EPS<=b && b<=c+EPS; }

//------------------------------------------------------------
// outer_prod
//   O
//------------------------------------------------------------

inline double outer_prod( const dcomp& a, const dcomp& b )
{
	return a.real()*b.imag() - a.imag()*b.real();
}

//----------------------------------------------------------------
// area_of_polygon
//   ʐ
//----------------------------------------------------------------

inline double area_of_polygon( const polygon& pts )
{
	// pts[0], pts[i], pts[i+1] ̂ȂOp`̖ʐςSB
	double acc = 0.0;
	if( pts.size() >= 3 )
		for( size_t i=1; i!=pts.size()-1; ++i )
			acc += outer_prod( pts[i]-pts[0], pts[i+1]-pts[i] );
	return abs(acc) / 2.0;
}

//----------------------------------------------------------------
// is_point_in_polygon
//   _p`̒ɂ邩H 2: 1:, 0:O
//----------------------------------------------------------------

inline int is_point_in_polygon( const dcomp& p, const polygon& poly )
{
	if( poly.empty() )
		return 0;

	// _p_ɗ悤ɕsړ
	polygon v( poly );
	for_each( v.begin(), v.end(), phoenix::arg1-p );
	v.push_back(v.front()); // ԕ

	// ӂxƂ̌񐔂JEgB
	int count = 0;
	for( size_t i=0; i!=poly.size(); ++i )
	{
		dcomp a=v[i], b=v[i+1];

		// abxƂ̌_vZc
		// a + s(b-a) = (x,0)
		// s = -a.y / (b.y-a.y)
		// x = a.x + s (b.x - a.x)

		double fy = b.imag() - a.imag();
		if( fy == 0.0 ) // abƕsƂ
		{
			if( is_zero(a.imag()) )
				if( is_ordered(a.real(),0,b.real())
				 || is_ordered(b.real(),0,a.real()) )
					return 1;
		}
		else // słȂƂ
		{
			double s = -a.imag() / fy;
			double x = a.real() + s * (b.real()-a.real());
			if( is_zero(x) ) // x_xƌ
			{
				if( is_ordered(a.real(),0,b.real())
				 || is_ordered(b.real(),0,a.real()) )
					return 1;
				return 0;
			}
			if( 0 < x ) // xŌ
			{
				// _X̂Ƃ́A̕ӂ""̒_ł鎞̂ݐ
				if( 0+EPS<=s && s<=1-EPS )                 ++count;
				else if( is_zero(s) && a.imag()<b.imag() ) ++count;
				else if( is_one (s) && b.imag()<a.imag() ) ++count;
			}
		}
	}

	// œO
	return (count%2==1 ? 2 : 0);
}

//----------------------------------------------------------------
// line_line_cross_pt
//   n_AsA̒ƁAn_BsB̒Ƃ̌_
//   sȏꍇ̗O͂Ȃ
//----------------------------------------------------------------

inline dcomp line_line_cross_pt( dcomp As, dcomp A, dcomp Bs, dcomp B )
{
	const dcomp S = Bs-As;
	double f = outer_prod(A, B);
	double q = outer_prod(S, A) / f;
	return Bs + q * B;
}

//----------------------------------------------------------------
// 
//----------------------------------------------------------------

inline bool LineSegLineSegIntersectCheck(
  const dcomp& As, const dcomp& Ae,
  const dcomp& Bs, const dcomp& Be
  )
{
  // O=[As] Bs-Ae-Be ̉] d1, d2 ̕ƂĎ擾
  dcomp a_Bs = Bs - As;
  dcomp a_Ae = Ae - As;
  dcomp a_Be = Be - As;
  double d1 = outer_prod( a_Bs, a_Ae );
  double d2 = outer_prod( a_Ae, a_Be );
  if( d1*d2 < 0 )
    return false;

  // O=[Bs] Ae-Be-As ̉] d3, d4 ̕ƂĎ擾
  dcomp b_Ae = Ae - Bs;
  dcomp b_Be = Be - Bs;
  dcomp b_As = As - Bs;
  double d3 = outer_prod( b_Ae, b_Be );
  double d4 = outer_prod( b_Be, b_As );
  if( d3*d4 < 0 )
    return false;

  if( d1*d3 > 0 )
    return true;
  if( is_zero(d1) || is_zero(d2) || is_zero(d3) || is_zero(d4) )
    return true; // [_܂ޏꍇ

  return false;
}
